// scan line converter
// Copyright 1997 Herbert Duerr

#ifndef FREETYPE

#include "xfstt.h"
#ifdef WIN32
	#include <windows.h>
	#include <memory.h>
#else
	#define __KERNEL__
	#include <asm/byteorder.h>
	#undef __KERNEL__
#endif

void Rasterizer::drawContour( int startPoint, int endPoint)
{
	register point *pp2, *pp3;
	int xpos, ypos;
	int xend, yend;

	int count = endPoint - startPoint;
	pp2 = &p[ 1][ startPoint];
	pp3 = pp2 + 1;

	if( pp2->flags) {
		xpos = xend = pp2->xnow;
		ypos = yend = pp2->ynow;
	} else if( pp3->flags) {
		xpos = xend = pp3->xnow;
		ypos = yend = pp3->ynow;
		--count;
	} else {
		xpos = xend = (pp2->xnow + pp3->xnow) >> 1;
		ypos = yend = (pp2->ynow + pp3->ynow) >> 1;
	}

	while( --count >= 0) {
		pp2 = pp3;
		pp3 = pp2 + 1;
		if( pp2->flags) {
			bezier0( xpos, ypos, pp2->xnow, pp2->ynow);
			xpos = pp2->xnow;
			ypos = pp2->ynow;
			continue;
		}
		if( count <= 0) {
			pp3 = &p[ 1][ startPoint];
			if( !pp3->flags) {
				int xtmp = (pp2->xnow + pp3->xnow) >> 1;
				int ytmp = (pp2->ynow + pp3->ynow) >> 1;
				bezier1( xpos, ypos, pp2->xnow, pp2->ynow, xtmp, ytmp);
				xpos = xtmp;
				ypos = ytmp;
				pp2 = pp3;
			}
			bezier1( xpos, ypos, pp2->xnow, pp2->ynow, xend, yend);
			return;
		}
		if( pp3->flags) {
			bezier1( xpos, ypos, pp2->xnow, pp2->ynow, pp3->xnow, pp3->ynow);
			xpos = pp3->xnow;
			ypos = pp3->ynow;
			++pp3;
			--count;
		} else {
			int xtmp = (pp2->xnow + pp3->xnow) >> 1;
			int ytmp = (pp2->ynow + pp3->ynow) >> 1;
			bezier1( xpos, ypos, pp2->xnow, pp2->ynow, xtmp, ytmp);
			xpos = xtmp;
			ypos = ytmp;
		}
	}
	pp2 = &p[ 1][ startPoint];
	if( pp2->flags)
		bezier0( xpos, ypos, xend, yend);
	else
		bezier1( xpos, ypos, pp2->xnow, pp2->ynow, xend, yend);
}


// the rendering is done by a quick and dirty XOR algorithm, which
// doesn't know a thing about dropout control

#if 0
void Rasterizer::drawSegment( int x1, int y1, int x2, int y2)
{
//	while( (y2-y1)>31 || (y1-y2)>31) {
	while( (y2-y1)>7 || (y1-y2)>7) {
		int x3 = (x1+x2)>>1;
		int y3 = (y1+y2)>>1;
		drawSegment( x1, y1, x3, y3);
		x1 = x3;
		y1 = y3;
	}

	// the algorithm here is a bit unconventional ...
	// basically it does only scan converter rule 1

	if( ((y1^y2) >> 5) != 1)
		return;

	char* ptr = &currBMP[ (height - 1 - (y1 >> SHIFT)) * dX];
	x1 = (x1 + 0x20) >> (SHIFT + 1);
	while( (x1 -= SCANLINEPAD) >= 0)
		(*ptr = ~*ptr), ++ptr;
	*ptr ^= (-1) << (-x1);
}
#else
inline void Rasterizer::drawSegment( int x1, int y1, int x2, int y2)
{
	if( y1 > y2) {
		int t=y1; y1=y2; y2=t;
		t=x1; x1=x2; x2=t;
	}

	int ygoal = (y2 & ~0x3F) + 0x20;
	if( ygoal > y2) ygoal -= 0x40;

	for( int x10 = x1, y10 = y1;
		ygoal > y1;
		x1 = x10, y1 = y10, ygoal -= 0x40) {

		x2 = x1 + (int)((float)(x2 - x1) * (ygoal - y1) / (y2 - y1));
		y2 = ygoal;

#if (LOGSLP == 3)
		U8* ptr = (U8*)&currBMP[ (height - 1 - (y2 >> SHIFT)) * dX];
#elif (LOGSLP == 5)
		U32* ptr = (U32*)&currBMP[ (height - 1 - (y2 >> SHIFT)) * dX];
#else
	#error "incorrect SCANLINEPAD"
#endif
		int t = (x2 + 0x20) >> SHIFT;
		if( t == 0)
			continue;
		while( (t -= SCANLINEPAD) > 0)
			(*ptr = ~*ptr), ++ptr;
#if (LOGSLP == 3)
		*ptr ^= (-1) << (-t);
#elif (LOGSLP == 5)
		*ptr ^= __ntohl((-1) << (-t));
#else
	#error "incorrect SCANLINEPAD"
#endif
	}
}
#endif


void Rasterizer::bezier1( int x00, int y00,
	int x01, int y01, int x02, int y02)
{
manual_tail_recursion:
	int dx = x02 - x00;
	int dy = y02 - y00;
	dx ^= (dx>>31);
	dy ^= (dy>>31);

	if( dx+dy < SUBS) {
		//bezier0( x00, y00, x02, y02);

		if( MAGNIFY == 0) {
			drawSegment( x00, y00, x02, y02);
			return;
		}
#if (LOGSLP == 3)
		U8* ptr = &currBMP[ (height - 1 - (y00 >> SHIFT)) * dX];
		ptr[ x00 >> (SHIFT+3)] |= 0x80 >> ((x00>>SHIFT) & 7);
#elif (LOGSLP == 5)
		U32* ptr = (U32*)&currBMP[ (height - 1 - (y00 >> SHIFT)) * dX];
		ptr[ x00 >> 11] |= __ntohl(0x80000000 >> ((x00>>SHIFT) & 31));
#else
	#error "Wrong LOGSLP"
#endif
		return;
	}

	int x10 = x00 + x01;
	int x11 = x01 + x02;
	int x20 = (x10 + x11) >> 2;
	x10 >>= 1;
	x11 >>= 1;

	int y10 = y00 + y01;
	int y11 = y01 + y02;
	int y20 = (y10 + y11) >> 2;
	y10 >>= 1;
	y11 >>= 1;

	bezier1( x00, y00, x10, y10, x20, y20);
	x00 = x20;	y00 = y20;
	x01 = x11;	y01 = y11;
	goto manual_tail_recursion;
}


// Bresenham algorithm
void Rasterizer::bezier0( int x00, int y00, int x01, int y01)
{
	if( MAGNIFY == 0) {
		drawSegment( x00, y00, x01, y01);
		return;
	}

	if( y01 < y00) {
		int t=y00; y00=y01; y01=t;
		t=x00; x00=x01; x01=t;
	}

	int dy0 = y01 - y00;
	int dx0 = x01 - x00;

#if (LOGSLP == 3)
	U8* ptr = &currBMP[ (height - 1 - (y00 >> SHIFT)) * dX];
#elif (LOGSLP == 5)
	U32* ptr = (U32*)&currBMP[ (height - 1 - (y00 >> SHIFT)) * dX];
#endif
	if( (U8*)ptr < currBMP || (U8*)ptr >= endBMP)
		return;
	if( (y01 >> SHIFT) >= (height - 1))
		y01 = (height - 1) << SHIFT;
	y01 -= y00 & ~0x3F;
	if( dx0 >= 0) {
		int d = dx0 - dy0;
		do {
			do {
#if (LOGSLP == 3)
				ptr[ x00 >> (SHIFT+3)] |= 0x80 >> ((x00>>SHIFT) & 7);
#elif (LOGSLP == 5)
				ptr[ x00 >> 11] |= __ntohl(0x80000000 >> ((x00>>SHIFT) & 31));
#endif
				if( d <= 0) break;
				d -= dy0;
				x00 += 64;
			} while( x00 <= x01);
			d += dx0;
			if( d >= 0) { d -= dy0; x00 += 64;}
			ptr -= dX>>2;
			y01 -= 64;
		} while( y01 > 0);
	} else {
		dx0 = -dx0;
		int d = dx0 - dy0;
		do {
			do {
#if (LOGSLP == 3)
				ptr[ x00 >> (SHIFT+3)] |= 0x80 >> ((x00>>SHIFT) & 7);
#elif (LOGSLP == 5)
				ptr[ x00 >> 11] |= __ntohl(0x80000000 >> ((x00>>SHIFT) & 31));
#endif
				if( d <= 0) break;
				d -= dy0;
				x00 -= 64;
			} while( x00 >= x01);
			d += dx0;
			if( d >= 0) { d -= dy0; x00 -= 64;}
			ptr -= dX>>2;
			y01 -= 64;
		} while( y01 > 0);
	}
}


// the secrets of "anti aliasing technology" ...

void Rasterizer::antiAliasing2()
{
	U8* p1 = currBMP + ((height * width + 3) >> 2);
	U8* p2 = p1 + dX;
	U8* p3 = currBMP;

	memcpy( p1, currBMP, length);

	for( int y = height>>1; --y >= 0;) {
		for( int x = dX; --x >= 0;) {
			U8 c1 = *(p1++);
			U8 c2 = *(p2++);
			for( int i = 4; --i >= 0;) {
				U8 c3 = (c1 & 1) + (c2 & 1);
				c3 += ((c1 & 2) + (c2 & 2)) >> 1;
				p3[ i] = c3;
				c1 >>= 2;
				c2 >>= 2;
			}
			p3 += 4;
		}
		p1 = p2;
		p2 = p1 + dX;
	}

	height >>= 1;
	width = dX * 4;
	dX = width;
	length = dX * height;
}

#endif /* FREETYPE */

