// Reading the TTFont
// Copyright 1997 Herbert Duerr

#include "xfstt.h"
#include <stdio.h>
#include <string.h>

TTFont::TTFont( char* fileName, int infoOnly):
	RandomAccessFile( fileName),
	nameTable( 0),	headTable( 0),	maxpTable( 0),
	cmapTable( 0),	locaTable( 0),	glyphTable( 0),
	fpgmTable( 0),	prepTable( 0),	cvtTable( 0),
	hheaTable( 0),	hmtxTable( 0),	
	os2Table( 0),
	ltshTable( 0),	hdmxTable( 0),	vdmxTable( 0),
	gaspTable( 0),	kernTable( 0)
	//postTable( 0)
{
	dprintf1( "TTFont( \"%s\");\n", fileName);

	/*U32 version		= */readUInt();
	short nTables		= readSShort();
	/*U16 searchRange	= */readUShort();
	/*U16 entrySelector	= */readUShort();
	/*U16 rangeShift	= */readUShort();

	for( int i = nTables; --i >= 0;) {
		int name	= readUInt();
		/*int checksum	= */readUInt();
		int offset	= readUInt();
		int length	= readUInt();
		//dprintf1( "checksum should be %08X\n", checksum);
		if( infoOnly && name != NAME_MAGIC && name != OS2_MAGIC)
			continue;
		switch( name) {
		case NAME_MAGIC:
			nameTable = new NameTable( *this, offset, length);
			break;
		case HEAD_MAGIC:
			headTable = new HeadTable( *this, offset, length);
			break;
		case MAXP_MAGIC:
			maxpTable = new MaxpTable( *this, offset, length);
			break;
		case CMAP_MAGIC:
			cmapTable = new CmapTable( *this, offset, length);
			break;
		case LOCA_MAGIC:
			locaTable = new LocaTable( *this, offset, length);
			break;
		case GLYF_MAGIC:
			glyphTable = new GlyphTable( *this, offset, length);
			break;
		case LTSH_MAGIC:
			ltshTable = new LtshTable( *this, offset, length);
			break;
		case OS2_MAGIC:
			os2Table = new OS2Table( *this, offset, length);
			break;
		case VDMX_MAGIC:
			vdmxTable = new VdmxTable( *this, offset, length);
			break;
		case CVT_MAGIC:
			cvtTable = new CvtTable( *this, offset, length);
			break;
		case FPGM_MAGIC:
			fpgmTable = new FpgmTable( *this, offset, length);
			break;
		case PREP_MAGIC:
			prepTable = new PrepTable( *this, offset, length);
			break;
		case HDMX_MAGIC:
			hdmxTable = new HdmxTable( *this, offset, length);
			break;
		case HHEA_MAGIC:
			hheaTable = new HheaTable( *this, offset, length);
			break;
		case HMTX_MAGIC:
			hmtxTable = new HmtxTable( *this, offset, length);
			break;
		case KERN_MAGIC:
			kernTable = new KernTable( *this, offset, length);
			break;
		case POST_MAGIC:
			//postTable = new PostTable( *this, offset, length);
			break;
		case GASP_MAGIC:
			gaspTable = new GaspTable( *this, offset, length);
			break;
		default:
			// System.out.println( "???");
			break;
		}
	}

	if( !(nameTable && headTable && maxpTable
	&& cmapTable && locaTable && glyphTable)) {
		dprintf0( "Incomplete TT file\n");
		return;
	}

	//int dummy;
	//nameTable->getString( 1, 3, &dummy);
	locaTable->setShortFlag( headTable->shortLoca());
	hmtxTable->setupHmtx( hheaTable->nLongHMetrics);
}


#ifndef WIN32
TTFont::~TTFont()
{
	closeRAFile();

	if( nameTable)	delete nameTable;
	if( headTable)	delete headTable;
	if( maxpTable)	delete maxpTable;
	if( cmapTable)	delete cmapTable;
	if( locaTable)	delete locaTable;
	if( glyphTable)	delete glyphTable;

	if( os2Table)	delete os2Table;
	if( cvtTable)	delete cvtTable;
	if( fpgmTable)	delete fpgmTable;
	if( prepTable)	delete prepTable;

	if( hheaTable)	delete hheaTable;
	if( hmtxTable)	delete hmtxTable;

	if( ltshTable)	delete ltshTable;
	if( hdmxTable)	delete hdmxTable;
	if( vdmxTable)	delete vdmxTable;

	if( gaspTable)	delete gaspTable;
	if( kernTable)	delete kernTable;
}
#endif

int TTFont::checkFont()
{
	return (headTable && headTable->checkHeadMagic());
} 

int TTFont::getEmUnits()
{
	return headTable->emUnits;
}

void TTFont::getFontInfo( FontInfo* fi)
{
	fi->firstChar	= os2Table->firstCharNo;
	fi->lastChar	= os2Table->lastCharNo;

	for( int i = 0; i < 10; ++i)
		fi->panose[ i] = os2Table->panose[ i];
	char* faceName = nameTable->getString( 1, 4, &fi->faceLength);
	if( !faceName) {
		//### e.g. Marlett doesn't have a 1.4 name
		faceName = "unknown";
		fi->faceLength = strlen( faceName);
	}
	if( fi->faceLength > 32)
		fi->faceLength = 32;
	strncpy( fi->faceName, faceName, fi->faceLength);
}

int TTFont::getGlyphData( int glyphNo, GlyphTable* g)
{
	if( glyphNo >= maxpTable->numGlyphs || glyphNo < 0)
		glyphNo = 0;
	int glyphOffset = locaTable->getGlyphOffset( glyphNo);
	if( glyphOffset < 0)	// empty glyph
		return 0;
	g->getGlyphData( glyphOffset, this);
	return 1;
}

int TTFont::getGlyphNo8( char char8)
{
	return cmapTable->char2glyphNo( char8);
}

int TTFont::getGlyphNo16( int unicode)
{
	return cmapTable->unicode2glyphNo( unicode);
}

#if 0
int TTFont::doCharNo8( GlyphTable* g, int char8)
{
	return doGlyphNo( g, getGlyphNo8( char8));
}

int TTFont::doCharNo16( GlyphTable* g, int unicode)
{
	return doGlyphNo( g, getGlyphNo16( unicode));
}
#endif

int TTFont::getMaxWidth( int mppemx)
{
	int maxWidth = 0;
	if( hdmxTable)
		maxWidth = hdmxTable->getMaxWidth( mppemx);
	if( maxWidth == 0) {
		maxWidth = headTable->xmax - headTable->xmin;
		maxWidth += headTable->emUnits >> 5;	// +3%
		maxWidth = maxWidth * mppemx / headTable->emUnits;
		dprintf1( "using maxWidth %d instead\n", maxWidth);
	}
	return maxWidth;
}

int TTFont::getGlyphWidth( int mppemx, int glyphNo)
{
	int width =  0;
	if( hdmxTable)
		width = hdmxTable->getGlyphWidth( mppemx, glyphNo);
	if( width == 0) {
		int dummy;
		hmtxTable->getHMetrics( glyphNo, &width, &dummy);
		//###if( width == 0)
		//###	width = getMaxWidth( mppemx):
		width += headTable->emUnits >> 5;	// +3%
		width = width * mppemx / headTable->emUnits;
		dprintf1( "using width %d instead\n", width);
	}
	return width;
}


// verify on reference implementation

void TTFont::patchGlyphCode( int patchFrom, int patchTo)
{
	int len = patchTo - patchFrom + 1;
	for( U8* ptr = base + patchFrom; --len >= 0; ++ptr)
		*ptr = CLEAR;
}

int TTFont::checksum( U8* buf, int len)
{
	len = (len + 3) >> 2;
	int sum = 0;
	for( U8* p = buf; --len >= 0; p += 4) {
		int val = (p[0]<<24) + (p[1]<<16) + (p[2]<<8) + p[3];
		sum += val;
	}
	return sum;
}

void TTFont::updateChecksums()
{
	U8* buf = base;

	int nTables = (buf[4] << 8) + buf[5];
	//printf( "nTables = %d\n", nTables);
	U8* headTable = 0;
	for( int i = 0; i < nTables; ++i) {
		U8* b = &buf[ 12 + i * 16];
		int name = (b[0]<<24) + (b[1]<<16) + (b[2]<<8) + b[3];
		int offset = (b[8]<<24) + (b[9]<<16) + (b[10]<<8) + b[11];
		int length = (b[12]<<24) + (b[13]<<16) + (b[14]<<8) + b[15];
		//printf( "offset = %08X, length = %08X\n", offset, length);
		int check = checksum( buf + offset, length);
		b[4] = (U8) (check >> 24);
		b[5] = (U8) (check >> 16);
		b[6] = (U8) (check >> 8);
		b[7] = (U8) check;
		//printf( "checksum[%d] = %08X\n", i, check);
		if( name == 0x68656164) {
			headTable = buf + offset;
			//printf( "headOffset = %08X\n", offset);
		}
	}

	int check = checksum( buf, getLength()) - 0xB1B0AFBA;
	//printf( "csAdjust = %08X\n", check);
	headTable[8] = (U8) (check >> 24);
	headTable[9] = (U8) (check >> 16);
	headTable[10] = (U8) (check >> 8);
	headTable[11] = (U8) check;
}

int TTFont::write2File( char* filename)
{
	FILE* fd = fopen( filename, "wb");
	if( fd) {
		int len = fwrite( base, 1, getLength(), fd);
		fclose( fd);
		return len;
	}
	return -1;
}


#if 0
#include <string.h>

/ xlfd has to be preset with the category name "-category-",
// returns "-category-family-weight-slant-setwidth-TT-"
//         "pixelsize-pointsize-xres-yres-spacing-avgwidth-charset-encoding"

int TTFont::getXLFDname( char* xlfd, int maxlen)
{
	char* p;

	int lenFamily;
	const char* strFamily	= nameTable->getString( 1, 1, &lenFamily);
	int lenSubFamily;
	const char* strSubFamily= nameTable->getString( 1, 2, &lenSubFamily);
	int lenUnique;
	const char* strUnique	= nameTable->getString( 1, 3, &lenUnique);

	if( strFamily == 0 || strSubFamily == 0 || strUnique == 0) {
		strFamily	= "Symbol";
		strSubFamily	= "SubFamily";
		strUnique	= "Unique";
		lenFamily	= strlen( strFamily);
		lenSubFamily	= strlen( strSubFamily);
		lenUnique	= strlen( strUnique);
	}

	int xlfdLen = strlen( xlfd);

	p = (char*)memmem( (__ptr_t)strSubFamily, lenSubFamily, (__ptr_t)"Bold", 4);
	char* weight = p ? "-bold" : "-medium";
	xlfdLen += strlen( weight);

	p = (char*)memmem( (__ptr_t)strSubFamily, lenSubFamily, (__ptr_t)"Italic", 6);
	char* slant = p ? "-i" : "-r";
	xlfdLen += strlen( slant);

	xlfdLen += strlen( "-normal-TT-0-0-0-0-m-0");
	xlfdLen += strlen( "iso8859-1") + 4;

	if( xlfdLen >= maxlen)
		return 0;

	strcat( xlfd, "-");
	strcat( xlfd, strFamily);
	strcat( xlfd, weight);
	strcat( xlfd, slant);
	strcat( xlfd, "-normal-TT-0-0-0-0-m-0");
	strcat( xlfd, "-iso8859-1");

	dprintf1( "xlfd = \"%s\"\n", xlfd);
	return xlfdLen;
}
#endif

