// Verify Character Outlines
// Copyright 1997 Herbert Duerr

#define BUFSIZE (1024*1024)

#include "xfstt.h"
#include <stdio.h>


static int fontSize = 9;

#ifndef WIN32
	#define TTFORIG		"//dosc/windows/fonts/cour.ttf"
#else
	//#define TTFORIG	"d:\\mysrc\\xfstt\\verdana.ttf"
	#define TTFORIG		"c:\\windows\\fonts\\cour.ttf"
	//#define TTFORIG	"c:\\windows\\fonts\\times.ttf"
	#define TTFNAME		"c:\\windows\\fonts\\sample.ttf"
#endif

#define VERBOSE 0

#ifndef MAGNIFY
	int MAGNIFY = 0;
#endif /*MAGNIFY*/

static int nPoints;
static point* points;
static int pointNumber;
static int* endPoints;
static int contourNumber;

void doPFX( U8* buf)
{
	U16 xfract = *(U16*) &buf[0];
	S16 xvalue = *(S16*) &buf[2];
	U16 yfract = *(U16*) &buf[4];
	S16 yvalue = *(S16*) &buf[6];

	int x = (xvalue << 6) + (xfract >> 10);
	int y = (yvalue << 6) + (yfract >> 10);

#if VERBOSE
	dprintf3( "\t\t\tPFX[%d] %d %d\n", pointNumber, x, y);
#endif

	for(; pointNumber < nPoints;){
		points[ pointNumber].xgoal = x;
		points[ pointNumber].ygoal = y;
		++pointNumber;
		if( points[ pointNumber-1].xnow != points[ pointNumber].xnow
		||  points[ pointNumber-1].ynow != points[ pointNumber].ynow)
			break;
	}
}

void doContour( U8* buf, int lenPoly)
{
	do {
#if VERBOSE
		char* typestr;
		U16 type = *(U16*) &buf[0];
		if( type == 1) typestr = "LINES";
		else if( type == 2) typestr = "CURVE";
		else typestr = "?????";
		dprintf1( "\t%s\n", typestr);
#endif

		U16 npts = *(U16*) &buf[2];
		buf += 4;
		for( int i = 0; i < npts; ++i, buf += 8)
			doPFX( buf);

		lenPoly -= 4 + npts * 8;
	} while( lenPoly > 0);
}

void doOutline( U8* buf, int lenChar)
{
	pointNumber = 0;
	for( contourNumber = 0; lenChar > 0; ++contourNumber) {
		U32 lenPoly = *(U32*) &buf[0];
		//U32 type = *(U32*) &buf[4];

		//dprintf1( "\tPOLY length = %d\n", lenPoly);
		//dprintf1( "\tPOLY type = %d (TTPOLYHEADER=24)\n", type);

		doPFX( buf + 8);
		doContour( buf + 16, lenPoly - 16);

		//### we only get one point if two successive points are equal
		pointNumber = endPoints[ contourNumber] + 1;
		buf += lenPoly;
		lenChar -= lenPoly;
	}

	point* pp = &points[ pointNumber];
	pp->xgoal = pp->ygoal = 0;
	++pp;
	pp->xgoal = pp->ygoal = 0;
}

	
#ifndef WIN32

int doFont( int argc, char**argv)
{
	char* ttfName;
	if( argc == 2) ttfName = argv[1];
	else ttfName = TTFORIG;

	TTFont* ttFont = new TTFont( ttfName);

	U8* buf = (U8*)allocMem( BUFSIZE);

	Rasterizer* raster = new Rasterizer();
	raster->useTTFont( ttFont);
	raster->setTransformation( fontSize, 0, 0, fontSize, 96, 96);

	for( int unicode = 0;;) {
		unicode = ttFont->cmapTable->nextUnicode( unicode);
		if( !unicode)
			break;

		GlyphTable* g = ttFont->glyphTable;
		int glyphNo = ttFont->getGlyphNo16( unicode);
		printf( "\nChar = \'%c\' -> %d\n", (U8)unicode, glyphNo);

		if( !ttFont->getGlyphData( glyphNo, g))
			continue;
		raster->scaleGlyph( g, glyphNo);

		nPoints		= g->nPoints;
		points		= g->points;
		endPoints	= g->endPoints;

		raster->printOutline();
	}

	deallocMem( (void*)buf, BUFSIZE);
	return 0;
}


int main( int argc, char**argv)
{
	return doFont( argc, argv);
}


#else

// no need for *.otl files when running on reference implementation

#include <windows.h>

//#ifdef DEBUG
FILE* outfile;
//#endif

int CALLBACK EnumFontFamProc(
    ENUMLOGFONT* lpelf, NEWTEXTMETRIC*  lpntm, int  fonttype, LPARAM  lParam)
{
	if( fonttype == TRUETYPE_FONTTYPE)
		dprintf1( "font: %s\n", lpelf->elfLogFont.lfFaceName);
	return 1;
}

LRESULT CALLBACK WndProc( HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	static TTFont* ttFont = NULL;
	static unsigned unicode = 0;
	static Rasterizer* raster = NULL;
	static U8* buf = NULL;
	char fontfamily[128];
	HDC	hdc;
	HGDIOBJ	oldfont, newfont;
	PAINTSTRUCT ps;
	long errsum;

	switch( msg) {
	case WM_CREATE:

		buf = (U8*)allocMem( BUFSIZE);
		break;

	case WM_PAINT:

	if( ttFont == NULL) {
		ttFont = new TTFont( TTFORIG);
		if( 0 >= ttFont->write2File( TTFNAME)) {
			dprintf1( "Cannot write file %s\n", TTFNAME);
			goto goodbye;
		}

		int namelen = 0;
		char* nameptr = ttFont->nameTable->getString( 1, 1, &namelen);
		strncpy( fontfamily, nameptr, namelen);
		fontfamily[ namelen] = '\0';
		dprintf1( "Font Family = %s\n", fontfamily);

		int i = CreateScalableFontResource( 0, "sample.fot", TTFNAME, NULL);
		i = AddFontResource( "sample.fot");

		raster = new Rasterizer();
		raster->useTTFont( ttFont);
		int resX = 72;	//GetDeviceCaps( hdc, LOGPIXELSX);
		int resY = 72;	//GetDeviceCaps( hdc, LOGPIXELSY);
		dprintf2( "resX = %d, resY = %d\n", resX, resY);
		raster->setTransformation( fontSize, 0, 0, fontSize, resX, resY);
	}

	{
		hdc = BeginPaint( hwnd, &ps);

		EnumFontFamilies( hdc, NULL, (FONTENUMPROC)EnumFontFamProc, 0);

		newfont = CreateFont( -fontSize, 0, 0, 0,
			FW_NORMAL, FALSE, FALSE, FALSE,
			ANSI_CHARSET, OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, DRAFT_QUALITY,
			VARIABLE_PITCH | FF_DONTCARE, fontfamily);
		oldfont = SelectObject( hdc, newfont);

		int size = GetFontData( hdc, 0, 0, NULL, 0);
		dprintf1( "fontsize = %d\n", size);
	}

	dprintf1( "MapMode = %d\n", GetMapMode( hdc));

	errsum = 0;
	for( unicode = 0; unicode <= 127;) {
		unicode = ttFont->cmapTable->nextUnicode( unicode);
		if( !unicode)
			break;

		U16 wch = (U16)unicode;
		TextOutW( hdc, 10, 10, &wch, 1);

		GlyphTable* g = ttFont->glyphTable;
		int glyphNo = ttFont->getGlyphNo16( unicode);
		fprintf( outfile, "\nChar = \'%C\'(%d) -> %d\n", unicode, unicode, glyphNo);

		if( !ttFont->getGlyphData( glyphNo, g))
			continue;
		raster->scaleGlyph( g, glyphNo);

		points		= g->points;
		endPoints	= g->endPoints;
		nPoints		= g->nPoints;

		MAT2 mat2 = {{0,1},{0,0},{0,0},{0,1}};
		GLYPHMETRICS gm;
		int charLen = GetGlyphOutline( hdc, unicode, GGO_NATIVE, &gm, 0, NULL, &mat2);

		if( charLen > 0 && charLen <= BUFSIZE) {
			GetGlyphOutline( hdc, unicode, GGO_NATIVE, &gm, charLen, buf, &mat2);
			doOutline( buf, charLen);
			++pointNumber;
			points[ pointNumber].xgoal = gm.gmCellIncX<<SHIFT;
			points[ pointNumber].ygoal = gm.gmCellIncY<<SHIFT;
		}

		raster->printOutline();
		dprintf2( "refX = %ld, refY = %ld\n",
			64*gm.gmptGlyphOrigin.x, 64*gm.gmptGlyphOrigin.y);
		dprintf2( "incX = %d, incY = %d\n", 64*gm.gmCellIncX, 64*gm.gmCellIncY);

		for( int i = 0; i < nPoints; ++i) {
			int xerr = points[i].xnow - points[i].xgoal;
			int yerr = points[i].ynow - points[i].ygoal;
			if( xerr < 0) xerr = -xerr;
			if( yerr < 0) yerr = -yerr;
			errsum += xerr + yerr;
		}
		dprintf2( "errsum[%c] = %ld\n", unicode, errsum);
	}

	dprintf1( "errsum = %ld\n", errsum);

	SelectObject( hdc, oldfont);
	DeleteObject( newfont);
	EndPaint( hwnd, &ps);
//	break;

	case WM_DESTROY:
goodbye:
		RemoveFontResource( "sample.fot");
		deallocMem( (void*)buf, BUFSIZE);
		PostQuitMessage( 0);
		break;

	default:
		return DefWindowProc( hwnd, msg, wparam, lparam);
		break;
    }
    return 0;
}

int WINAPI WinMain( HINSTANCE hInst, HINSTANCE hPrev, LPSTR cmd, int nCmdShow)
{
//#ifdef DEBUG
	outfile = fopen( "trace", "wb");
//#endif

	WNDCLASS wndclass;
	wndclass.style	= CS_GLOBALCLASS | CS_PARENTDC | CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc	= WndProc;
	wndclass.cbClsExtra	= 0;
	wndclass.cbWndExtra	= 0;
	wndclass.hInstance	= hInst;
	wndclass.hIcon		= NULL;
	wndclass.hCursor	= NULL;
	wndclass.hbrBackground	= GetStockObject( WHITE_BRUSH);
	wndclass.lpszMenuName	= NULL;
	wndclass.lpszClassName	= "TTVERIFY";
	RegisterClass( &wndclass);

	HWND hwnd = CreateWindow( "TTVERIFY", "TT Verifier",
	    WS_OVERLAPPED,
	    CW_USEDEFAULT, CW_USEDEFAULT,
	    250, 250,
	    NULL, NULL, hInst, NULL);

	ShowWindow( hwnd, nCmdShow);
	UpdateWindow( hwnd);

	MSG msg;
	while( GetMessage( &msg, NULL, 0, 0)) {
	    TranslateMessage( &msg);
	    DispatchMessage( &msg);
	}

//#ifdef DEBUG
	fclose( outfile);
//#endif
	return 0;
}

#endif

