// Show Character Outlines for Debugging Purposes
// Copyright 1997 Herbert Duerr

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdio.h>

#include "xfstt.h"

static Display* display;
static Visual* visual;
static int screen;
static int depth;
static Colormap colorMap;
static Window rootWindow;

static GC textGC;
static int color[5];
static int yellow;
static int black;
static int white;

#define STARTGLYF 6
//#define DEFAULT_FONT	"/DOS/windows/system/couri.ttf"
//#define DEFAULT_FONT	"../myjava/ttfedit/fonts/anastas.ttf"
#define DEFAULT_FONT	"verdana.ttf"

static Rasterizer* raster;
static U8* pixmap;

static int fontsize = 32;
static XImage* img = 0;

#define BMPSIZE 2048*2048
U8* bitmap = 0;

#define DEFAULT_WIDTH	200
#define DEFAULT_HEIGHT	220

#ifndef MAGNIFY
	int MAGNIFY = 1;
#endif /*MAGNIFY*/

XImage* glyph2image( int charNo, int sz)
{
	int length;
	static int old_font_size = 0;

	if( !bitmap)
		bitmap = (U8*)allocMem( BMPSIZE);

	if( sz != old_font_size ) {
		raster->setTransformation( sz, 0, 0, sz, 96, 96);
		printf( "fontsize = %d\n", sz);
	}
	old_font_size = sz;
	GlyphMetrics gm;
	length = raster->putGlyphBitmap( charNo, BMPSIZE, bitmap, &gm);

	if( length == 0)
		return img;

	if( raster->anti_aliasing > 0)
		for( int i = length; --i >= 0;)
			pixmap[i] = color[ bitmap[i]];
#ifndef MAGNIFY
	else if( MAGNIFY) {
		S8* p1 = (S8*)bitmap;
		U8* p2 = pixmap;
		raster->dX *= 8;	// 1 bpp -> 8 bpp
		for( int h = 0; h < raster->height; ++h) {
			int cy = h / MAGNIFY;
			S8 m;
			for( int w = 0; w < raster->dX; ++w) {
				if( (w & 7) == 0)
					m = *(p1++);
				int c = color[4];
				if( m >= 0)
					c = color[ ((w / MAGNIFY) ^ cy) & 1];
				*(p2++) = c;
				m <<= 1;
			}
		}
	}
#endif

	if( !img)
		img = new XImage();

	int isPixmap = (MAGNIFY || raster->anti_aliasing > 0);

	img->width		= raster->width;
	img->height		= raster->height;
	img->xoffset		= 0;
	img->format		= isPixmap ? ZPixmap : XYBitmap;
	img->data		= (char*)(isPixmap ? pixmap : bitmap);
	img->byte_order		= LSBFirst;
	img->bitmap_unit	= 8;
	img->bitmap_bit_order	= MSBFirst;
	img->bitmap_pad		= 0;
	img->depth		= isPixmap ? 8 : 1;
	img->bytes_per_line	= raster->dX;
	img->bits_per_pixel	= isPixmap ? 8 : 1;
	img->red_mask		= 0xFF0000;
	img->green_mask		= 0x00FF00;
	img->blue_mask		= 0x0000FF;
	img->obdata		= 0;

	return img;
}

int xAllocColor( int rgb)
{
	XColor c;
	c.red = (rgb >> 8) & 0xff00;
	c.green = rgb & 0xff00;
	c.blue = (rgb << 8) & 0xff00;
	XAllocColor( display, colorMap, &c);
	return c.pixel;
}

int main( int argc, char** argv)
{
	if( argc > 2) {
		fprintf( stderr, "Usage: show.e [filename]\nthe default ");
		fprintf( stderr, "filename is \"%s\"\n", DEFAULT_FONT);
		return -1;
	}

	char* ttFileName = (argc == 2) ? argv[1] : DEFAULT_FONT;

	pixmap = new U8[ 1024 * 1024];

	int done = 0;

	display = XOpenDisplay( ":0");
	visual = XDefaultVisual( display, screen);
	screen = XDefaultScreen( display);
	depth = XDefaultDepth( display, screen);
	rootWindow = XDefaultRootWindow( display);
	colorMap = XDefaultColormap( display, screen);

	color[0] = xAllocColor( 0xF0F0F0);
	color[1] = xAllocColor( 0xB4B4B4);
	color[2] = xAllocColor( 0x787878);
	color[3] = xAllocColor( 0x3C3C3C);
	color[4] = xAllocColor( 0x000000);

	white = xAllocColor( 0xF0F0F0);
	black = xAllocColor( 0x000000);
	yellow = xAllocColor( 0xFFFF00);

	XGCValues gcv;
	gcv.function = GXcopy;
	gcv.foreground = black;
	gcv.background = white;
	gcv.line_width = 1;
	gcv.line_style = LineSolid;
	unsigned long valuemask = GCFunction | GCForeground;
	textGC = XCreateGC( display, rootWindow, valuemask, &gcv);

	unsigned int width = DEFAULT_WIDTH, height = DEFAULT_HEIGHT;

	Window win = XCreateSimpleWindow( display, rootWindow, 0, 0,
		width, height, 1, gcv.foreground, gcv.background);
	XStoreName( display, win, "TrueType Viewer");

	int selection_mask = StructureNotifyMask | ExposureMask
	//	| EnterWindowMask | LeaveWindowMask
		| ButtonPressMask | ButtonReleaseMask
	//	| PointerMotionMask
		| KeyPressMask | KeyReleaseMask;
	XSelectInput( display, win, selection_mask);

	XMapWindow( display, win);

	raster = new Rasterizer();
	TTFont* ttFont = new TTFont( ttFileName);
	raster->useTTFont( ttFont);

	XImage* img = 0;

	XEvent event;
	do
	{
		static int glyphNo = STARTGLYF;
		XNextEvent( display, &event);
		switch( event.type)
		{
drawglyph:
#ifndef MAGNIFY
			if( MAGNIFY && fontsize) {
				MAGNIFY = height / (fontsize + 1);
				if( !MAGNIFY) MAGNIFY = 1;
			}
#endif
			img = glyph2image( glyphNo, fontsize);
			/* fall through */
		case Expose:
expose:
			XSetForeground( display, textGC, yellow);
			XFillRectangle( display, win, textGC, 0, 0, width, height);
			XSetForeground( display, textGC, black);
			if( img)
				XPutImage( display, win, textGC,
					img, 0, 0, 4, 4,
					img->width, img->height);
			break;

		case KeyPress:
		{
			char c[8];
			unsigned long ks;
			XComposeStatus xcs;
			XLookupString( &event.xkey, c, 8, &ks, &xcs);
			switch( ks) {
			case XK_Escape:
				done = 1;
				break;
			case XK_F2:
				raster->grid_fitting = !raster->grid_fitting;
				goto drawglyph;
			case XK_F3:
				raster->anti_aliasing = !raster->anti_aliasing;
				goto drawglyph;
			case XK_F4:
#ifndef MAGNIFY
				if( MAGNIFY)
					MAGNIFY = 0;
				else
					MAGNIFY = 1;
				goto drawglyph;
#endif /*MAGNIFY*/
			case XK_Left:
				--glyphNo;
				goto drawglyph;
			case XK_Right:
				++glyphNo;
				goto drawglyph;
			case XK_Up:
				fontsize += 1;
				goto drawglyph;
			case XK_Down:
				fontsize -= 1;
				if( fontsize < 4)
					fontsize = 4;
				goto drawglyph;
			case XK_Page_Up:
				fontsize += (fontsize >> 2) ? fontsize >> 2 : 1;
				goto drawglyph;
			case XK_Page_Down:
				fontsize -= (fontsize >> 2) ? fontsize >> 2 : 1;
				goto drawglyph;
			case XK_Begin:
				fontsize >>= 1;
				if( fontsize < 4)
					fontsize = 4;
#ifndef MAGNIFY
				else
					MAGNIFY <<= 1;
#endif /*MAGNIFY*/
				goto drawglyph;
			case XK_End:
				fontsize <<= 1;
#ifndef MAGNIFY
				MAGNIFY >>= 1;
#endif /*MAGNIFY*/
				goto drawglyph;
			default:
				printf( "key = \"%s\"\n", c);
				glyphNo = ttFont->getGlyphNo16( c[0]);
				img = glyph2image( glyphNo, fontsize);
			}
			goto expose;
		}

		case ConfigureNotify:
			width = event.xconfigure.width;
			height = event.xconfigure.height;
			goto drawglyph;
			//break;

		case DestroyNotify:
			done = 1;
			break;

		default: 
			//printf( "XEvent.type = %d\n", event.type);
			break;
		}
	} while( !done);

	if( bitmap)
		deallocMem( bitmap, BMPSIZE);

	delete raster;
	delete ttFont;
	delete[] pixmap;
	delete[] img;

	cleanupMem();

	XUnmapWindow( display, win);
	XCloseDisplay( display);
	return 0;
}

