// X Font Server for *.ttf Files
// Copyright 1997 Herbert Duerr

//#undef DEBUG
//#define DEBUG 1

//### why doesn't MAXFONTS 8 or 16 work?
#define MAXFONTS 63

#define MAXREPLYSIZE 256000

#include "xfstt.h"
#include "ttfn.h"

#include <stdio.h>
#include <dirent.h>
#include <strings.h>

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <X11/fonts/FS.h>
#include <X11/fonts/FSproto.h>
#include <X11/Xlib.h>

struct XFSFont
{
	Font		fid;
	TTFont*		ttFont;
	FontInfo	fi;
	FontExtent	fe;
} xfsFont[ MAXFONTS];

#ifndef MAGNIFY
	int MAGNIFY = 0;
#endif /* MAGNIFY */

static int fdTtfnFile, fdNameFile;
static int lenTtfnFile, lenNameFile;
static U8 *ptrTtfnFile, *ptrNameFile;

void usage()
{
	printf( "Usage: xfstt [--sync]\n");
} 

int ttSyncDir( FILE* infoFile, FILE* nameFile, char* ttdir)
{
	int nfonts = 0;
	printf( "xfstt sync in directory \"%s\"\n", ttdir);
	DIR* dirp = opendir( ".");

	while( dirent* de = readdir( dirp)) {
		int namelen = strlen( de->d_name);
		if( namelen - 4 <= 0) continue;
		char* ext = &de->d_name[ namelen - 4];
		if( ext[0] != '.') continue;
		if( (ext[1] | 0x20) != 't') continue;
		if( (ext[2] | 0x20) != 't') continue;
		if( (ext[3] | 0x20) != 'f') continue;

		struct stat statbuf;
		stat( de->d_name, &statbuf);
		if( !S_ISREG( statbuf.st_mode))
			continue;

		char pathName[ 256];
		strcpy( pathName, ttdir);
		strcat( pathName, "/");
		strcat( pathName, de->d_name);

		TTFont* ttFont = new TTFont( de->d_name, 1);
		FontInfo fi;
		ttFont->getFontInfo( &fi);

		TTFNdata info;
		info.nameOfs		= ftell( nameFile);
		info.nameLen		= fi.faceLength;
		info.pathLen		= strlen( pathName);

		fwrite( (void*)fi.faceName, 1, info.nameLen, nameFile);
		fputc( '\0', nameFile);
		fwrite( (void*)pathName, 1, info.pathLen, nameFile);
		fputc( '\0', nameFile);

		info.charSet		= 'U';
		info.slant		= 0;
		info.bFamilyType	= fi.panose[ 0];
		info.bSerifStyle	= fi.panose[ 1];
		info.bWeight		= fi.panose[ 2];
		info.bProportion	= fi.panose[ 3];
		info.bContrast		= fi.panose[ 4];
		info.bStrokeVariation	= fi.panose[ 5];
		info.bArmStyle		= fi.panose[ 6];
		info.bLetterForm	= fi.panose[ 7];
		info.bMidLine		= fi.panose[ 8];
		info.bXHeight		= fi.panose[ 9];

		fwrite( (void*)&info, 1, sizeof( info), infoFile);

		delete ttFont;
		++nfonts;
	}

	closedir( dirp);
	return nfonts;
}

int ttSyncAll()
{
	dprintf0( "TT synching\n");

	chdir( "/etc/ttfonts");
	DIR* dirp = opendir( ".");
	FILE* infoFile = fopen( "ttinfo.dir", "wb");
	FILE* nameFile = fopen( "ttname.dir", "wb");
	if( infoFile <= 0 || nameFile <= 0) {
		dprintf0( "TT synch failed\n");
		return -1;
	}

	TTFNfile info = { {'T','T','F','N'}, {'I','N','F','O'}, 0x0100};
	fwrite( (void*)&info, 1, sizeof( info), infoFile);
	strncpy( info.type, "NAME", 4);
	fwrite( (void*)&info, 1, sizeof( info), nameFile);

	int nfonts = 0;
	chdir( "/etc/ttfonts");
	nfonts += ttSyncDir( infoFile, nameFile, ".");
	while( dirent* de = readdir( dirp)) {
		chdir( "/etc/ttfonts");
		if( de->d_name[0] == '.' || chdir( de->d_name) != 0)
			continue;
		nfonts += ttSyncDir( infoFile, nameFile, de->d_name);
	}

	fclose( infoFile);
	fclose( nameFile);

	return nfonts;
}


int listFonts( char* pattern, int firstFlag, char* buf)
{
	static TTFNdata* ttfn = 0;

	if( pattern[ 0] != '*' || pattern[ 1] != 0)
		return 0;

	if( firstFlag || ttfn == 0)
		ttfn = (TTFNdata*)(ptrTtfnFile + sizeof( TTFNfile));

	if( (U8*)ttfn >= ptrTtfnFile + lenTtfnFile)
		return 0;

	char* fontName = (char*)ptrNameFile + ttfn->nameOfs;

	TPFontName fn;
	sprintf( &fn.panose[0][0],
		"%02X%02X%02X%02X%02X"
		"%02X%02X%02X%02X%02X",
		ttfn->bFamilyType,
		ttfn->bSerifStyle,
		ttfn->bWeight,
		ttfn->bProportion,
		ttfn->bContrast,
		ttfn->bStrokeVariation,
		ttfn->bArmStyle,
		ttfn->bLetterForm,
		ttfn->bMidLine,
		ttfn->bXHeight);

	fn.nameLen	= sizeof( TPFontName) + ttfn->nameLen;
	fn.magic[ 0]	= 'T';
	fn.magic[ 1]	= 'T';
	fn.charset	= ttfn->charSet;
	fn.modifier	= '0';
	fn.panoseMagic	= 'P';

	strncpy( buf, (char*)&fn, sizeof( fn));
	strncpy( buf + sizeof( fn), fontName, ttfn->nameLen);

	buf[ fn.nameLen] = 0;
	dprintf1( "ListFont \"%s\"\n", buf);

	++ttfn;
	return fn.nameLen+1;
}

int openFont( Rasterizer* raster,
	char* ttfnName, int size, int xres, int yres, int fid)
{
	if( ttfnName[0] != 'T' || ttfnName[1] != 'T')
		return -1;

	// parse attributes

	int midx = 0, ridx = 0;
	int mm[4], rr[2], flags = 0;

	for( ttfnName += 2;;) {
		char c = *(ttfnName++);
		if( c == '\0') return -1;
		if( c == '_') break;
		int neg = 0;
		if( c == '-') { ++ttfnName, neg = -1;}
		int val = 0;
		while( *ttfnName >= '0' && *ttfnName <= '9')
			val = 10*val + *(ttfnName++) - '0';
		if( neg)
			val = -val;
		switch( c) {
		case 'M':
			if( midx < 4) mm[ midx++] = val;
			break;
		case 'R':
			if( ridx < 2) rr[ ridx++] = val;
			break;
		case 'F':
			flags = val;
			break;
		default:
			return -1;
		}
	}

	// set attribute defaults

	switch( midx) {
	case 0:	mm[0] = size;		//fall through
	case 1:	mm[1] = mm[0];	//fall through
	case 2:	mm[2] = 0;		//fall through
	case 3:	mm[3] = -mm[2];	//fall through
	case 4:	break;
	}

	switch( ridx) {
	case 0: rr[0] = xres;	//fall through
	case 1: rr[1] = yres;	//fall through
	case 2: break;
	}

	// open font file

	chdir( "/etc/ttfonts");

	TTFont* ttFont = 0;
	TTFNdata* ttfn = (TTFNdata*)( ptrTtfnFile + sizeof( TTFNfile));
	while( (U8*) ttfn < ptrTtfnFile + lenTtfnFile) {
		char* name = (char*)( ptrNameFile + ttfn->nameOfs);
		if( !strcmp( name, ttfnName)) {
			name += ttfn->nameLen + 1;
			ttFont = new TTFont( name);
			break;
		}
		++ttfn;
	}

	if( !ttFont)
		return -1;

	// init rasterizer

	int fontIndex;
	for( fontIndex = 0; fontIndex < MAXFONTS; ++fontIndex)
		if( xfsFont[ fontIndex].fid == 0)
			break;
	if( fontIndex >= MAXFONTS){
		delete ttFont;
		return -1;
	}

	xfsFont[ fontIndex].fid		= fid;
	xfsFont[ fontIndex].ttFont	= ttFont;
	ttFont->getFontInfo( &xfsFont[ fontIndex].fi);

	//### hack to prevent XFree crashes
	FontInfo* fi = &xfsFont[ fontIndex].fi;
	if( fi->lastChar > 0xff) fi->lastChar = 0xff;
	if( fi->firstChar > fi->lastChar) fi->firstChar = fi->lastChar;

	raster->useTTFont( ttFont, flags);
	raster->setTransformation( mm[0], mm[2], mm[3], mm[1], rr[0], rr[1]);

	FontExtent* fe	= &xfsFont[ fontIndex].fe;
	int numGlyphs	= ttFont->maxpTable->getNumGlyphs();
	fe->buflen	= MAXREPLYSIZE;
	fe->buflen	+= numGlyphs * sizeof( CharInfo) + sizeof(void*);
	fe->buffer	= (U8*)allocMem( fe->buflen);
	raster->getFontExtent( fe);

	return fontIndex;
}

int findFont( Font fid, int fd, int seqno)
{
	for( int i = 0; i < MAXFONTS; ++i) 
		if( xfsFont[ i].fid == fid)
			return i;

#if 1
	//### return valid font index for now
	dprintf1( "fid = %d not found!\n", fid);
#else
	fsError reply;
	reply.type		= FS_Error;
	reply.request		= 2;
	reply.sequenceNumber	= seqno;
	reply.length		= sizeof(reply) >> 2;
	write( fd, (void*)&reply, sizeof(reply));
#endif
	return -1;
}
	
#define MAXREQSIZE	4096
char buf[MAXREQSIZE];

int connecting( int fd)
{
	int i = read( fd, buf, MAXREQSIZE);	// read fsConnClientPrefix

	fsConnClientPrefix* req = (fsConnClientPrefix*) buf;
	if( i < (int)sizeof(fsConnClientPrefix))
		return -1;

	dprintf0( "Connecting\n");
	dprintf1( "%s endian connection\n",
		(req->byteOrder == 'l') ? "little" : "big");
	dprintf2( "version %d.%d\n", req->major_version, req->minor_version);

	fsConnSetup replySetup;
	replySetup.status		= 0;
	replySetup.major_version	= 2;
	replySetup.minor_version	= 0;
	replySetup.num_alternates	= 0;
	replySetup.auth_index		= 0;
	replySetup.alternate_len	= 0;
	replySetup.auth_len		= 0;

	write( fd, (void*)&replySetup, sizeof( replySetup));

	struct {
		fsConnSetupAccept s1;
		char vendor[2];
		char pad[2];
	} replyAccept;
	replyAccept.s1.length		= (sizeof( replyAccept) + 3) >> 2;
	replyAccept.s1.max_request_len	= MAXREQSIZE;
	replyAccept.s1.vendor_len	= sizeof( replyAccept.vendor);
	replyAccept.s1.release_number	= 0;
	replyAccept.vendor[0]		= 'H';
	replyAccept.vendor[1]		= 'D';

	write( fd, (void*)&replyAccept, sizeof( replyAccept));
	return 0;
}

void bitswap( U8* p, int len)
{
	// oh no... this makes me sick
	U8 t4[ 16] = {0,8,4,12,2,10,6,14,1,9,5,13,3,11,7,15};
	for( ; --len >= 0; ++p)
		*p = (t4[ *p & 15] << 4) | t4[ (*p >> 4) & 15];
}

int working( int fd)
{
	int event_mask = 0;
	int xres = 96, yres = 96, size = 18;

	struct stat statbuf;
	stat( "/etc/ttfonts/ttinfo.dir", &statbuf);
	lenTtfnFile	= statbuf.st_size;
	fdTtfnFile	= open( "/etc/ttfonts/ttinfo.dir", O_RDONLY);
	ptrTtfnFile	= (U8*)mmap( 0L, lenTtfnFile, PROT_READ, MAP_SHARED, fdTtfnFile, 0L);

	stat( "/etc/ttfonts/ttname.dir", &statbuf);
	lenNameFile	= statbuf.st_size;
	fdNameFile	= open( "/etc/ttfonts/ttname.dir", O_RDONLY);
	ptrNameFile	= (U8*)mmap( 0L, lenNameFile, PROT_READ, MAP_SHARED, fdNameFile, 0L);

	Rasterizer* raster = new Rasterizer();

	for( int seqno = 1;; ++seqno) {
		int l = read( fd, buf, sizeof(fsGenericReply));
		if( l < 4)
			break;

		int length = (buf[2] + (buf[3] << 8)) << 2;
		if( length > MAXREQSIZE) {
			dprintf1( "request too big: %d bytes\n", length);
			break;
		}

		while( l < length)
			l += read( fd, buf+l, length-l);

		switch( buf[0]) {	// request type
		case FS_Noop:
			dprintf0( "FS_Noop\n");
			break;

		case FS_ListExtensions:
			dprintf0( "###FS_ListExtensions\n");
			break;

		case FS_QueryExtension:
			dprintf0( "###FS_QueryExtension\n");
			break;

		case FS_ListCatalogues:
			dprintf0( "###FS_ListCatalogues\n");
			break;

		case FS_SetCatalogues:
			dprintf0( "###FS_SetCatalogues\n");
			break;

		case FS_GetCatalogues:
			dprintf0( "###FS_GetCatalogues\n");
			break;

		case FS_SetEventMask:
			{
			fsSetEventMaskReq* req = (fsSetEventMaskReq*) buf;
			event_mask = req->event_mask;
			dprintf1( "FS_SetEventMask %04X\n", event_mask);
			}
			break;

		case FS_GetEventMask:
			dprintf1( "FS_GetEventMask = %04X\n", event_mask);
			{
			fsGetEventMaskReply reply;
			reply.type		= FS_Reply;
			reply.sequenceNumber	= seqno;
			reply.length		= sizeof(reply) >> 2;
			reply.event_mask	= event_mask;
			write( fd, (void*)&reply, sizeof(reply));
			}
			break;

		case FS_CreateAC:
			dprintf0( "FS_CreateAC\n");
			{
			fsCreateACReply reply;
			reply.type		= FS_Reply;
			reply.auth_index	= 0;
			reply.sequenceNumber	= seqno;
			reply.length		= sizeof(reply) >> 2;
			reply.status		= 0;
			write( fd, (void*)&reply, sizeof(reply));
			}
			break;

		case FS_FreeAC:
			dprintf0( "FS_FreeAC\n");
			{
			fsGenericReply reply;
			reply.type		= FS_Reply;
			reply.sequenceNumber	= seqno;
			reply.length		= sizeof(reply) >> 2;
			write( fd, (void*)&reply, sizeof(reply));
			}
			break;

		case FS_SetAuthorization:
			dprintf0( "FS_SetAuthorization\n");
			break;

		case FS_SetResolution:
			{
			fsSetResolutionReq* req = (fsSetResolutionReq*) buf;
			int numres = req->num_resolutions;
			dprintf1( "FS_SetResolution * %d\n", numres);
			fsResolution* res = (fsResolution*) (buf+sizeof(req));
			for(; --numres >= 0; ++res) {
				dprintf3( "xres = %d, yres = %d, size = %d\n",
					xres = res->x_resolution,
					yres = res->y_resolution,
					size = res->point_size);
			}
			}
			break;

		case FS_GetResolution:
			dprintf0( "FS_GetResolution\n");
			{
			struct {
				fsGetResolutionReply s1;
				fsResolution s2;
				char pad[2];
			} reply;
			reply.s1.type		= FS_Reply;
			reply.s1.num_resolutions= 1;
			reply.s1.sequenceNumber	= seqno;
			reply.s1.length		= sizeof(reply) >> 2;
			reply.s2.x_resolution	= xres;
			reply.s2.y_resolution	= yres;
			reply.s2.point_size	= size;
			write( fd, (void*)&reply, sizeof(reply));
			}

			break;

		case FS_ListFonts:
			{
			fsListFontsReq* req = (fsListFontsReq*) buf;
			char* pattern = (char*) (req + 1);
			pattern[ (req+1)->nbytes] = 0;
			dprintf2( "FS_ListFonts \"%s\" * %d\n", pattern, req->maxNames);

			fsListFontsReply reply;
			reply.type		= FS_Reply;
			reply.sequenceNumber	= seqno;
			reply.following		= 0;	//### XFREE doesn't handle this concept
			reply.nFonts		= 0;

			char* buf0 = new char[ MAXREPLYSIZE];
			char* buf = buf0;
			for( U32 i = 0; i < req->maxNames; ++i) {
				int len = listFonts( pattern, i==0, buf);
				if( len == 0) break;
				++reply.nFonts;
				buf += len;
			}
			reply.length = sizeof( reply);
			reply.length = (reply.length + (buf - buf0) + 3) >> 2;
			write( fd, (void*)&reply, sizeof( reply));
			write( fd, (void*)buf0, (reply.length << 2) - sizeof(reply));
			delete[] buf0;
			}
			break;

		case FS_ListFontsWithXInfo:
			dprintf0( "FS_ListFontsWithXInfo\n");
			//### TODO
			break;

		case FS_OpenBitmapFont:
			{
			fsOpenBitmapFontReq* req = (fsOpenBitmapFontReq*) buf;
			char* fontName = (char*)(req + 1)+1;
			fontName[ *(U8*)(req+1)] = 0;
			dprintf1( "FS_OpenBitmapFont \"%s\"", fontName);

			if( 0>openFont( raster, fontName,
				size, xres, yres, req->fid)) {
				fsError reply;
				reply.type		= FS_Error;
				reply.request		= 7;
				reply.sequenceNumber	= seqno;
				reply.length		= sizeof(reply) >> 2;
				write( fd, (void*)&reply, sizeof(reply));
				dprintf0( " not found\n");
			} else {
				fsOpenBitmapFontReply reply;
				reply.type		= FS_Reply;
				reply.otherid_valid	= fsFalse;
				reply.sequenceNumber	= seqno;
				reply.length		= sizeof(reply) >> 2;
				reply.otherid		= 0;
				reply.cachable		= fsTrue;
				write( fd, (void*)&reply, sizeof(reply));
				dprintf0( " opened\n");
			}
			dprintf3( "fhint = %04lX, fmask = %04lX, fid = %d\n",
				req->format_hint, req->format_mask, req->fid);
			}
			break;

		case FS_QueryXInfo:
			{
			fsQueryXInfoReq* req = (fsQueryXInfoReq*) buf;
			dprintf1( "FS_QueryXInfo fid = %d\n", req->id);

			struct {
				fsQueryXInfoReply s1;
				fsPropInfo s2;
				fsPropOffset s3;
				U32 dummyName, dummyValue;
			} reply;
			reply.s1.type		= FS_Reply;
			reply.s1.sequenceNumber	= seqno;
			reply.s1.length		= sizeof(reply) >> 2;
			reply.s1.font_header_flags	= 0;

			int i = findFont( req->id, fd, seqno);
			if( i < 0)
				break;
			FontInfo* fi = &xfsFont[ i].fi;
			FontExtent* fe = &xfsFont[ i].fe;

			reply.s1.font_hdr_char_range_min_char_high	= (U8)(fi->firstChar>>8);
			reply.s1.font_hdr_char_range_min_char_low	= (U8)fi->firstChar;
			reply.s1.font_hdr_char_range_max_char_high	= (U8)(fi->lastChar>>8);
			reply.s1.font_hdr_char_range_max_char_low	= (U8)fi->lastChar;

			dprintf2( "minchar = 0x%02X%02X, ",
				reply.s1.font_hdr_char_range_min_char_high,
				reply.s1.font_hdr_char_range_min_char_low);
			dprintf2( "maxchar = 0x%02X%02X\n",
				reply.s1.font_hdr_char_range_max_char_high,
				reply.s1.font_hdr_char_range_max_char_low);

			reply.s1.font_header_draw_direction	= LeftToRightDrawDirection;
			reply.s1.font_header_default_char_high	= 0;
			reply.s1.font_header_default_char_low	= ' '; //###

			reply.s1.font_header_min_bounds_left	= fe->xLeftMin;
			reply.s1.font_header_min_bounds_right	= fe->xRightMin;
			reply.s1.font_header_min_bounds_width	= fe->xAdvanceMin;
			reply.s1.font_header_min_bounds_ascent	= fe->yAscentMin;
			reply.s1.font_header_min_bounds_descent	= fe->yDescentMin;
			reply.s1.font_header_min_bounds_attributes	= 0;
			reply.s1.font_header_max_bounds_left	= fe->xLeftMax;
			reply.s1.font_header_max_bounds_right	= fe->xRightMax;
			reply.s1.font_header_max_bounds_width	= fe->xAdvanceMax;
			reply.s1.font_header_max_bounds_ascent	= fe->yAscentMax;
			reply.s1.font_header_max_bounds_descent	= fe->yDescentMax;
			reply.s1.font_header_max_bounds_attributes	= 0;
			reply.s1.font_header_font_ascent	= fe->yWinAscent;
			reply.s1.font_header_font_descent	= fe->yWinDescent;

			// we need to have some property data, otherwise
			// the X server complains
			reply.s2.num_offsets	= 1;
			reply.s2.data_len	= 8;
			reply.s3.name.position	= 0;
			reply.s3.name.length	=
			reply.s3.value.position	= 4;
			reply.s3.value.length	= 4;
			reply.s3.type		= 0; //???
			reply.dummyName		= 0x414243;
			reply.dummyValue	= 0x636261;

			write( fd, (void*)&reply, sizeof(reply));
			}
			break;

		case FS_QueryXExtents8:
			//### QueryXExtents8 and QueryXExtents16 are different
			// fall through
		case FS_QueryXExtents16:
			{
			fsQueryXExtents8Req* req = (fsQueryXExtents8Req*) buf;
			//### req->format == 0x208
			dprintf2( "FS_QueryXExtents%s fid = %d, ",
				(buf[0] == FS_QueryXExtents16 ? "16" : "8"),
				req->fid);
			dprintf2( "range=%d, nranges=%ld\n",
				req->range, req->num_ranges);

			int i = findFont( req->fid, fd, seqno);
			if( i < 0)
				break;
			FontInfo* fi	= &xfsFont[ i].fi;
			FontExtent* fe	= &xfsFont[ i].fe;
			TTFont* font	= xfsFont[ i].ttFont;

			CharInfo* ci	= (CharInfo*)( fe->buffer + 4);

			fsQueryXExtents8Reply reply;
			reply.type		= FS_Reply;
			reply.sequenceNumber	= seqno;
			reply.length		= sizeof( reply) >> 2;
			reply.num_extents	= fi->lastChar - fi->firstChar + 1;
			//### ...
			fsXCharInfo* ext = new fsXCharInfo[ reply.num_extents];
			reply.length += (reply.num_extents * sizeof(*ext) + 3) >> 2;

			for( int ch = fi->firstChar; ch <= fi->lastChar; ++ch) {
				int j = ch - fi->firstChar;
				int glyphNo = font->getGlyphNo16( ch);
				GlyphMetrics* gm = &ci[ glyphNo].gm;
				
				ext[ j].left		= -gm->xOrigin;
				ext[ j].right		= gm->xBlackbox - gm->xOrigin;
				ext[ j].width		= gm->xAdvance;
				ext[ j].ascent		= gm->yBlackbox - gm->yOrigin;
				ext[ j].descent		= gm->yOrigin;
				ext[ j].attributes	= 0;

				dprintf3( "GM[%c] = (l= %d, r= %d, ", ch, ext[j].left, ext[j].right);
				dprintf3( "w= %d, a= %d, d= %d);\n", ext[j].width, ext[j].ascent, ext[j].descent);
			}
			write( fd, (void*)&reply, sizeof( reply));
			write( fd, (void*)ext, reply.num_extents * sizeof( *ext));
			delete[] ext;
			}
			break;

		case FS_QueryXBitmaps8:
			//### QueryXBitmaps8 and 16 are different
			// fall through
		case FS_QueryXBitmaps16:
			{
			fsQueryXBitmaps16Req* req = (fsQueryXBitmaps16Req*) buf;
			dprintf2( "FS_QueryXBitmaps16 fid = %d, fmt = %04lX\n",
				req->fid, req->format);

			int i = findFont( req->fid, fd, seqno);
			if( i < 0)
				break;
			FontInfo* fi	= &xfsFont[ i].fi;
			FontExtent* fe	= &xfsFont[ i].fe;
			TTFont* font	= xfsFont[ i].ttFont;

			CharInfo* ci	= (CharInfo*)( fe->buffer + 4);

			fsQueryXBitmaps16Reply reply;
			reply.type		= FS_Reply;
			reply.sequenceNumber	= seqno;
			reply.length		= sizeof(reply) >> 2;
			reply.num_chars = fi->lastChar - fi->firstChar + 1;
			reply.nbytes		= fe->bmplen;
			reply.replies_hint	= 0;
			reply.length += (reply.num_chars * 8 + reply.nbytes) >> 2;

			fsOffset32* ofs = new fsOffset32[ reply.num_chars];
			fsOffset32* o = ofs;
			for( int ch = fi->firstChar; ch <= fi->lastChar; ++ch) {
				int glyphNo	= font->getGlyphNo16( ch);
				o->position	= ci[ glyphNo].offset;
				o->length	= ci[ glyphNo].length;

				dprintf3( "OFS[%04X = %c] = %d\n", ch, ch, o->position);

				++o;
			}
			write( fd, (void*)&reply, sizeof(reply));
			write( fd, (void*)ofs, reply.num_chars * sizeof(*ofs));
			delete[] ofs;

			void* bitmaps = (void*)ci;
			bitmaps	+= (*(U32*)fe->buffer) * sizeof(CharInfo);

			if( (req->format & BitmapFormatMaskBit) == 0)
				bitswap( (U8*)bitmaps, fe->bmplen);

			write( fd, bitmaps, reply.nbytes);
			}
			break;

		case FS_CloseFont:
			{
			fsCloseReq* req = (fsCloseReq*) buf;
			dprintf1( "FS_CloseFont fid = %d\n", req->id);

			int i = findFont( req->id, fd, seqno);
			if( i < 0)
				break;
			FontExtent* fe	= &xfsFont[ i].fe;
			deallocMem( fe->buffer, fe->buflen);
			xfsFont[ i].fid = 0;
			}
			break;

		default:
			dprintf1( "Unknown FS request %d !\n", buf[0]);
			{
			fsRequestError reply;
			reply.type		= FS_Error;
			reply.request		= buf[0];
			reply.sequenceNumber	= seqno;
			reply.length		= sizeof(reply) >> 2;
			reply.timestamp		= 0;
			reply.major_opcode	= buf[0];
			reply.minor_opcode	= buf[1];
			write( fd, (void*)&reply, sizeof(reply));
			}
			break;
		}
	dprintf0( "done.\n");
	}

	munmap( (void*)ptrTtfnFile, lenTtfnFile);
	munmap( (void*)ptrNameFile, lenNameFile);
	close( fdTtfnFile);
	close( fdNameFile);

	close( fd);
	return 0;
}


int main( int argc, char** argv)
{
	for( int i = 1; i < argc; ++i) {
		if( !strcmp( argv[i], "--sync")) {
			ttSyncAll();
			cleanupMem();
			return 0;
		}
		usage();
		return -1;
	}

	int sfd = socket( PF_UNIX, SOCK_STREAM, 0);

	int tmp[50];
	struct sockaddr *sa = (struct sockaddr*)tmp;
	sa->sa_family = AF_UNIX;
	strcpy( sa->sa_data, "/tmp/.font-unix/fs7100");
	int sa_len = 24;
	mkdir( "/tmp/.font-unix", 0777);
	unlink( sa->sa_data);
	bind( sfd, sa, sa_len);

	int status = listen( sfd, 1);	// only one connection
	dprintf1( "listen = %d\n", status);

	*(sa->sa_data) = 0;
	sa_len = 100;
	//int ffd = accept( sfd, sa, &sa_len);
	int ffd = accept( sfd, 0L, 0L);
	//dprintf2( "sa->data = %s, sa_len = %d\n", sa->sa_data, sa_len);
	dprintf1( "accept = %d\n", ffd);

	connecting( ffd);
	working( ffd);

	shutdown( sfd, 2);
	cleanupMem();
	return 0;
}

